import json
import re


class ExManager:

    with open('../regexes/expressions.txt') as fh:
        expressions = json.load(fh)

    def get_compiled(self):
        regexes = {k.lower(): re.compile(self.expressions[k]) for k, v in self.expressions.items()}
        return regexes

    def print_expressions(self):
        {print(k.lower(), ":", self.expressions[k]) for k, v in sorted(self.expressions.items())}



import argparse
import glob
import html
import os
import pickle
import re
import time
from collections import Counter
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy
from tqdm import tqdm

REGEX_TOKEN = re.compile(r'(?<![#@])\b[a-z]{1,15}\b')
REGEX_URL = re.compile(r"(https?:\/\/(?:www\.|(?!www))[^\s\.]+\.[^\s]{2,}|www\.[^\s]+\.[^\s]{2,})")
SEPARATOR = "_"


##############################################################################################################
# Parse Arguments
##############################################################################################################

def check_empty_arg(value):
    if len(str(value)) == 0:
        raise argparse.ArgumentTypeError("Invalid argument - no value passed.")
    return value


def parse_int_list(value):
    if len(str(value)) == 0:
        raise argparse.ArgumentTypeError("Invalid argument - no value passed.")

    return [int(x) for x in value.split(",")]


parser = argparse.ArgumentParser()

# add arguments ########################################
parser.add_argument('--input', nargs='?', type=check_empty_arg, default="./",
                    help='path to file or directory containing the files for calculating the statistics.')
parser.add_argument('--name', nargs='?', type=check_empty_arg, default="mycorpus", help='')
parser.add_argument('--ngrams', type=int, default=2, help='up-to how many ngrams to calculate statistics.')
parser.add_argument('--mincount', nargs='+', type=int, default=[60, 25],
                    help='eliminate all ngrams below the given count.')
parser.add_argument('--perc', nargs='+', type=int, default=0,
                    help='eliminate all ngrams below the given percentile. 0=ALL')

pickle_parser = parser.add_mutually_exclusive_group()
pickle_parser.add_argument('--pickle', dest='pickle', action='store_true')
pickle_parser.add_argument('--no-pickle', dest='pickle', action='store_false')
parser.set_defaults(pickle=False)

web_parser = parser.add_mutually_exclusive_group()
web_parser.add_argument('--web-fix', dest='web_fix', action='store_true')
web_parser.add_argument('--no-web-fix', dest='web_fix', action='store_false')
parser.set_defaults(web_fix=True)

args = parser.parse_args()

args.input = "C:\\Users\\theos\\Downloads\\text8\\"
args.name = "text8"


print(args)


##############################################################################################################

def tokenize(text):
    """
    extract words from text
    :param text:
    :return:
    """
    if args.web_fix:
        text = REGEX_URL.sub(' ', text)
        text = html.unescape(text)
    return REGEX_TOKEN.findall(text.lower())


def get_ngrams(input_list, n):
    return zip(*[input_list[i:] for i in range(n)])


def write_stats_to_file(filename, counts, mincount):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename + ".txt", 'w', encoding="utf-8") as f:
        if args.perc == 0:
            percentile = 0
        else:
            percentile = numpy.percentile(numpy.fromiter(counts.values(), numpy.int32), args.perc)
        threshold = max(percentile, mincount)

        for k, v in counts.items():

            if v >= threshold:
                entry = k.split(SEPARATOR)
                entry.append(str(v))
                f.write('\t'.join(entry) + '\n')

    if args.pickle:
        with open(filename + ".pickle", 'wb') as f:
            pickle.dump(counts, f)


def count_file(filename, countkeeper, desc=""):
    """
    Count the word statistics of a file
    :param desc:
    :param filename:
    :param countkeeper:
    :return:
    """
    print()
    print("computing statistics for file: ", filename)
    with open(filename, "r", encoding="utf-8", errors='ignore') as infile:
        num_lines = sum(1 for line in open(filename, "r", encoding="utf-8"))
        for line in tqdm(infile, total=num_lines, desc=desc):
            try:
                toks = tokenize(line)
                for i in range(args.ngrams):
                    ngram = i + 1
                    if ngram > 1:
                        toks = ["<S>"] + toks
                    for token in get_ngrams(toks, ngram):
                        countkeeper[ngram][SEPARATOR.join(token)] += 1
            except Exception as e:
                print("ERROR - ", e, infile)


def write_stats(counts):
    print()
    for k, v in counts.items():
        print("Writing " + str(k) + "-grams...")
        counter = Counter(counts[k])
        print("entries:{}\t-\ttokens:{}".format(format(len(counter), ','), format(sum(counter.values()), ',')))
        write_stats_to_file("../stats/" + args.name + "/counts_" + str(k) + "grams", counter, args.mincount[int(k) - 1])


def prune_low_freq(word_stats, threshold):
    # remove ngrams with count less than mincount
    # avoid dict comprehension as it creates a new temp dict and overloads the memory
    for ng in list(word_stats.keys()):
        for t in list(word_stats[ng].keys()):
            if not word_stats[ng][t] >= threshold:
                del word_stats[ng][t]


def plot_statistics(statistics):
    fig = plt.figure(figsize=(5 * len(statistics), 5))
    for i, (k, v) in enumerate(statistics.items()):
        ax = fig.add_subplot(1, len(statistics), i + 1)
        ax.set_title("{}-gram - total={}".format(k, len(v)))
        ax.grid(True)
        values = numpy.fromiter(statistics[k].values(), numpy.int32)
        ax.hist(values, bins=100, range=(0, 100))
    fig.tight_layout()
    fig.canvas.draw()
    fig.canvas.flush_events()


if __name__ == '__main__':
    plt.ion()  # set plot to animated
    stats = defaultdict(lambda: defaultdict(int))
    pruning_size_threshold = 5000000
    low_freq_threshold = 3

    if os.path.isfile(args.input):
        count_file(args.input, stats)
        time.sleep(0.01)
        prune_low_freq(stats, 1)
        write_stats(stats)

    elif os.path.isdir(args.input):
        files = glob.glob(args.input + "*.txt")
        for i, file in enumerate(files):
            try:
                count_file(file, stats, str(i + 1) + "/" + str(len(files)))
            except Exception as e:
                print("ERROR - ", e, file)

            time.sleep(0.01)

            if any(len(stats[ngram]) > pruning_size_threshold for ngram in list(stats.keys())):
                print("Cleaning entries with only one occurrence, to save memory...")
                prune_low_freq(stats, low_freq_threshold)
                # write progress
                # plot_statistics(stats)

            write_stats(stats)

        prune_low_freq(stats, low_freq_threshold)
        write_stats(stats)
    else:
        print("Wrong input. Give a file or directory!")



"""
Created by Christos Baziotis.
"""
import nltk

from ekphrasis.classes.tokenizer import SocialTokenizer, Tokenizer


def wsp_tokenizer(text):
    return text.split(" ")


puncttok = nltk.WordPunctTokenizer().tokenize

social_tokenizer = SocialTokenizer(lowercase=False).tokenize
mytokenizer = Tokenizer(lowercase=False).tokenize

sents = [
    "CANT WAIT for the new season of #TwinPeaks ＼(^o^)／ yaaaay!!! #davidlynch #tvseries :)))",
    "I saw the new #johndoe movie and it suuuuucks!!! WAISTED $10... #badmovies 3:/",
    "@SentimentSymp:  can't wait for the Nov 9 #Sentiment talks!  YAAAAAAY !!! >:-D http://sentimentsymposium.com/.",
]

for s in sents:
    print()
    print("ORG: ", s)  # original sentence
    # print("WSP : ", wsp_tokenizer(s))  # whitespace tokenizer
    # print("WPU : ", puncttok(s))  # WordPunct tokenizer
    print("SC : ", social_tokenizer(s))  # social tokenizer
    print("SC : ", mytokenizer(s))  # social tokenizer



from functools import lru_cache

import ftfy

from ekphrasis.classes.exmanager import ExManager
from ekphrasis.classes.segmenter import Segmenter
from ekphrasis.classes.spellcorrect import SpellCorrector
from ekphrasis.utils.nlp import unpack_contractions


# noinspection PyPackageRequirements
class TextPreProcessor:
    def __init__(self, **kwargs):
        """
        Kwargs:
            omit (list): choose what tokens that you want to omit from the text.
                possible values: ['email', 'percent', 'money', 'phone', 'user',
                    'time', 'url', 'date', 'hashtag']
                Important Notes:
                            1 - put url at front, if you plan to use it.
                                Messes with the regexes!
                            2 - if you use hashtag then unpack_hashtags will
                                automatically be set to False

            normalize (list): choose what tokens that you want to normalize
                from the text.
                possible values: ['email', 'percent', 'money', 'phone', 'user',
                    'time', 'url', 'date', 'hashtag']
                for example: myaddress@mysite.com will be transformed to <email>
                Important Notes:
                            1 - put url at front, if you plan to use it.
                                Messes with the regexes!
                            2 - if you use hashtag then unpack_hashtags will
                                automatically be set to False

            unpack_contractions (bool): Replace *English* contractions in
                ``text`` str with their unshortened forms
                for example: can't -> can not, wouldn't -> would not, and so on...

            unpack_hashtags (bool): split a hashtag to it's constituent words.
                for example: #ilikedogs -> i like dogs

            annotate (list): add special tags to special tokens.
                possible values: ['hashtag', 'allcaps', 'elongated', 'repeated']
                for example: myaddress@mysite.com -> myaddress@mysite.com <email>

            tokenizer (callable): callable function that accepts a string and
                returns a list of strings if no tokenizer is provided then
                the text will be tokenized on whitespace

            segmenter (str): define the statistics of what corpus you would
                like to use [english, twitter]

            corrector (str): define the statistics of what corpus you would
                like to use [english, twitter]

            spell_correct_elong (bool): choose if you want to perform
                spell correction after the normalization of elongated words.
                * significantly affects performance (speed)

            spell_correction (bool): choose if you want to perform
                spell correction to the text
                * significantly affects performance (speed)

            fix_text (bool): choose if you want to fix bad unicode terms and
                html entities.
        """
        self.omit = kwargs.get("omit", {})
        self.backoff = kwargs.get("normalize", {})
        self.include_tags = kwargs.get("annotate", {})
        self.unpack_contractions = kwargs.get("unpack_contractions", False)
        self.tokenizer = kwargs.get("tokenizer", None)
        self.dicts = kwargs.get("dicts", None)
        self.spell_correction = kwargs.get("spell_correction", False)
        self.spell_correct_elong = kwargs.get("spell_correct_elong", False)
        self.fix_text = kwargs.get("fix_bad_unicode", False)
        self.unpack_hashtags = kwargs.get("unpack_hashtags", False)
        self.segmenter_corpus = kwargs.get("segmenter", "english")
        self.corrector_corpus = kwargs.get("corrector", "english")
        self.mode = kwargs.get("mode", "normal")

        if self.unpack_hashtags:
            self.segmenter = Segmenter(corpus=self.segmenter_corpus)
        if self.mode != "fast":
            self.spell_corrector = SpellCorrector(corpus=self.corrector_corpus)

        self.regexes = ExManager().get_compiled()
        if 'hashtag' in self.omit or 'hashtag' in self.backoff:
            print("You can't omit/backoff and unpack hashtags!\n unpack_hashtags will be set to False")
            self.unpack_hashtags = False

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    @staticmethod
    def add_special_tag(m, tag, mode="single"):

        if isinstance(m, str):
            text = m
        else:
            text = m.group()

        if mode == "single":
            return " " + text + " " + "<" + tag + ">" + " "
        elif mode == "wrap":
            return " ".join([" <" + tag + "> ", text, " </" + tag + "> "]) + " "
        elif mode == "every":
            tokens = text.split()
            processed = " ".join([" " + t + " " + "<" + tag + ">" + " " for t in tokens])
            return " " + processed + " "

    @lru_cache(maxsize=131072)
    def handle_hashtag_match(self, m):
        """
        Break a string to its constituent words (using Viterbi algorithm)
        """
        text = m.group()[1:]

        # todo:simplify routine
        if text.islower():
            expanded = self.segmenter.segment(text)
            expanded = " ".join(expanded.split("-"))
            expanded = " ".join(expanded.split("_"))
            # print(m.group(), " - ", expanded)
            # with open("analysis/segmenter_" + self.segmenter_corpus + ".txt", "a") as f:
            #     f.write(m.group() + "\t" + expanded + "\n")

        else:
            # split words following CamelCase convention
            expanded = self.regexes["camel_split"].sub(r' \1', text)
            expanded = expanded.replace("-", "")
            expanded = expanded.replace("_", "")
            # print(m.group(), " - ", expanded)

        if "hashtag" in self.include_tags:
            expanded = self.add_special_tag(expanded, "hashtag", mode="wrap")

        return expanded

    def handle_elongated_match(self, m):
        text = m.group()

        # normalize to at most 2 repeating chars
        text = self.regexes["normalize_elong"].sub(r'\1\1', text)

        normalized = self.spell_corrector.normalize_elongated(text)
        if normalized:
            text = normalized

        # try to spell correct the word
        if self.spell_correct_elong:
            text = self.spell_corrector.correct_word(text, assume_wrong=True, fast=True)
            # with open("analysis/spell_corrector_" + self.corrector_corpus + ".txt", "a") as f:
            #     f.write(m.group() + " - " + text + "\n")

            # print(m.group(), "-", text)
        if "elongated" in self.include_tags:
            text = self.add_special_tag(text, "elongated")

        return text

    @lru_cache(maxsize=4096)
    def handle_repeated_puncts(self, m):
        """
        return the sorted set so mathes random combinations of puncts will be mapped to the same token
        "!??!?!!", "?!!!!?!", "!!?", "!?!?" --> "?!"
        "!...", "...?!" --> ".!"
        :param m:
        :return:
        """
        text = m.group()
        text = "".join(sorted(set(text), reverse=True))

        if "repeated" in self.include_tags:
            text = self.add_special_tag(text, "repeated")

        return text

    @lru_cache(maxsize=4096)
    def handle_generic_match(self, m, tag):
        """
        :param tag:
        :param m:
        :return:
        """
        text = m.group()
        if tag in self.include_tags:
            text = self.add_special_tag(text, tag, mode="every")

        return text

    @lru_cache(maxsize=4096)
    def handle_emphasis_match(self, m):
        """
        :param m:
        :return:
        """
        text = m.group().replace("*", "")
        if "emphasis" in self.include_tags:
            text = self.add_special_tag(text, "emphasis")

        return text

    @staticmethod
    def dict_replace(wordlist, _dict):
        return [_dict[w] if w in _dict else w for w in wordlist]

    @lru_cache(maxsize=131072)
    def pre_process_doc(self, doc):

        # ###########################
        # # fix bad unicode
        # ###########################
        # if self.fix_bad_unicode:
        #     doc = textacy.preprocess.fix_bad_unicode(doc)
        #
        # ###########################
        # # fix html leftovers
        # ###########################
        # doc = html.unescape(doc)

        ###########################
        # fix text
        ###########################
        if self.fix_text:
            doc = ftfy.fix_text(doc)

        ###########################
        # BACKOFF & OMIT
        ###########################
        for item in self.backoff:
            # better add an extra space after the match. Just to be safe. extra spaces will be normalized later anyway
            doc = self.regexes[item].sub(lambda m: " " + "<" + item + ">" + " ", doc)
        for item in self.omit:
            doc = doc.replace("<" + item + ">", '')

        ###########################
        # unpack hashtags
        ###########################
        if self.unpack_hashtags:
            doc = self.regexes["hashtag"].sub(lambda w: self.handle_hashtag_match(w), doc)

        ###########################
        # handle special cases
        ###########################
        if self.mode != "fast":
            doc = self.regexes["elongated"].sub(lambda w: self.handle_elongated_match(w), doc)
            doc = self.regexes["repeat_puncts"].sub(lambda w: self.handle_repeated_puncts(w), doc)
            doc = self.regexes["emphasis"].sub(lambda w: self.handle_emphasis_match(w), doc)
            doc = self.regexes["allcaps"].sub(lambda w: self.handle_generic_match(w, "allcaps"), doc)
            doc = self.regexes["censored"].sub(lambda w: self.handle_generic_match(w, "censored"), doc)

        ###########################
        # unpack contractions: i'm -> i am, can't -> can not...
        ###########################

        # remove textacy dependency
        if self.unpack_contractions:
            doc = unpack_contractions(doc)

        ###########################
        # Tokenize
        ###########################
        doc = " ".join(doc.split())  # normalize whitespace
        if self.tokenizer:
            doc = self.tokenizer(doc)

            # Replace tokens with special dictionaries (slang,emoticons ...)
            # todo: add spell check before!
            if self.dicts:
                for d in self.dicts:
                    doc = self.dict_replace(doc, d)

        return doc

    def pre_process_docs(self, docs):
        from tqdm import tqdm
        # for d in docs:
        for d in tqdm(docs, desc="PreProcessing..."):
            yield self.pre_process_doc(d)



import re
from functools import lru_cache
from math import log10

from ekphrasis.classes.exmanager import ExManager
from ekphrasis.utils.helpers import read_stats

"""
The Segmenter Class implements the Viterbi algorithm for word segmentation.
Based on CH14 from the book Beautiful Data (Segaran and Hammerbacher, 2009)
"""

REGEX_TOKEN = re.compile(r'\b[a-z]{2,}\b')
NGRAM_SEP = "_"  # todo: move to values


class Pdist(dict):
    """
    A probability distribution estimated from word counts
    Notice: if pw = Pdist(unigrams, n_tokens:
        * pw[w] is the raw count of the word w
        * pw(w) is the probability of the word w
    """

    @staticmethod
    def default_unk_func(key, total):
        return 1. / total

    def __init__(self, data=None, total=None, unk_func=None, **kwargs):
        super().__init__(**kwargs)

        # insert the word counts
        data = data or {}
        for key, count in data.items():
            self[key] = self.get(key, 0) + int(count)

        self.total = float(total or sum(self.values()))
        self.unk_prob = unk_func or self.default_unk_func

    def __call__(self, key):
        if key in self:
            return self[key] / self.total
        else:
            return self.unk_prob(key, self.total)


class Segmenter:
    def __init__(self, corpus="english", max_split_length=20):
        """
        Args:
            corpus (str): the statistics from which corpus to use for
                the spell correction.
            max_split_length (int): the maximum length of that a word can have
                for looking for splits
        """

        # self.unigrams = Counter(read_stats(corpus, 1))
        # self.bigrams = Counter(read_stats(corpus, 2))
        self.unigrams = read_stats(corpus, 1)
        self.bigrams = read_stats(corpus, 2)
        self.N = sum(self.unigrams.values())
        self.L = max_split_length

        self.Pw = Pdist(self.unigrams, self.N, self.unk_probability)
        self.P2w = Pdist(self.bigrams, self.N)

        self.case_split = ExManager().get_compiled()["camel_split"]

    def condProbWord(self, word, prev):
        """
        Conditional probability of word, given previous word
        if bigram is not in our list, then fall back to unigrams
        Args:
            word (): candidate word
            prev (): previous observed word

        Returns:

        """
        try:
            return self.P2w[prev + NGRAM_SEP + word] / float(self.Pw[prev])
        except KeyError:
            return self.Pw(word)

    @staticmethod
    def unk_probability(key, total):
        """
        Estimate the probability of an unknown word, penalizing its length
        :param key: the word
        :param total: the count of all tokens
        :return:
        """
        return 10. / (total * 10 ** len(key))

    @staticmethod
    def combine(first, rem):
        """
        Combine first and rem results into one (probability, words) pair
        :param first: a tuple in the form: probability, word
        :param rem: a tuple in the form: probability, list_of_words
        :return:
        """
        (first_prob, first_word) = first
        (rem_prob, rem_words) = rem
        return first_prob + rem_prob, [first_word] + rem_words

    def splits(self, text):
        """
        Return a list of all possible (first, rem) pairs with max length of first <=L
        :param text:
        :return:
        """
        return [(text[:i + 1], text[i + 1:])
                for i in range(min(len(text), self.L))]

    # if you don't have enough RAM lower the maxsize
    @lru_cache(maxsize=1048576)
    def find_segment(self, text, prev='<S>'):
        """
        Return (log P(words), words), where words is the best estimated segmentation
        :param text: the text to be segmented
        :param prev:
        :return:
        """
        if not text:
            return 0.0, []
        candidates = [self.combine((log10(self.condProbWord(first, prev)), first), self.find_segment(rem, first))
                      for first, rem in self.splits(text)]
        return max(candidates)

    # if you don't have enough RAM lower the maxsize
    @lru_cache(maxsize=524288)
    def segment(self, word):
        if word.islower():
            return " ".join(self.find_segment(word)[1])
        else:
            return self.case_split.sub(r' \1', word).lower()

    def demo(self):
        print("BBCtest: ", self.segment('BbcTest'))
        print("choosespain: ", self.segment('choosespain'))
        print("speedofart: ", self.segment('speedofart'))
        print("smallandinsignificant: ", self.segment('smallandinsignificant'))

# Segmenter().demo()



from setuptools import setup, find_packages

setup(name='ekphrasis',
      version='0.2.0',
      description='Text processing tool, geared towards text from social networks, such as Twitter or Facebook. Ekphrasis performs tokenization, word normalization, word segmentation (for splitting hashtags) and spell correction.',
      url='https://github.com/cbaziotis/ekphrasis',
      author='Christos Baziotis',
      author_email='christos.baziotis@gmail.com',
      license='MIT',
      packages=find_packages(exclude=['docs', 'tests*', 'analysis']),
      include_package_data=True
      )
